// Fill out your copyright notice in the Description page of Project Settings.


#include "PeelVCamFunctionLibrary.h"

#include "LevelSequence.h"
#include "MovieSceneSequence.h"
#include "MovieSceneTrack.h"
#include "MovieSceneObjectBindingID.h"
#include "Channels/MovieSceneChannelProxy.h"
#include "Engine/World.h"
#include "GameFramework/Actor.h"
#include "MovieScene.h"
#include "ISequencer.h"
#include "LevelSequenceActor.h"
#include "ILevelSequenceEditorToolkit.h"
#include "Modules/ModuleManager.h"
#include "Sections/MovieScene3DTransformSection.h"
#include "Channels/MovieSceneFloatChannel.h"
#include "ISequencerModule.h"
#include "MovieScene.h"
#include "MovieSceneSection.h"


void UPeelVCamFunctionLibrary::Refresh(ULevelSequence* TargetSequence)
{

    IAssetEditorInstance* AssetEditor = GEditor->GetEditorSubsystem<UAssetEditorSubsystem>()->FindEditorForAsset(TargetSequence , false);
    ILevelSequenceEditorToolkit* LevelSequenceEditor = static_cast<ILevelSequenceEditorToolkit*>(AssetEditor);
    TWeakPtr<ISequencer> WeakSequencer = LevelSequenceEditor ? LevelSequenceEditor->GetSequencer() : nullptr;

    if (WeakSequencer.IsValid())
    {
        WeakSequencer.Pin()->NotifyMovieSceneDataChanged(EMovieSceneDataChangeType::MovieSceneStructureItemsChanged);
    }
}
// Helper function to get the current frame time as FFrameTime
FFrameTime UPeelVCamFunctionLibrary::GetCurrentFrameTime(ULevelSequence* Sequence)
{
	FFrameRate rate = Sequence->GetMovieScene()->GetDisplayRate();
	int32 CurrentFrame = ULevelSequenceEditorBlueprintLibrary::GetCurrentLocalTime();
	return FFrameRate::TransformTime(CurrentFrame, rate, Sequence->GetMovieScene()->GetTickResolution());
}

// Helper function to update or add keys to transform channels
void UPeelVCamFunctionLibrary::UpdateOrAddKeysToTransformChannels(UMovieScene3DTransformSection* TransformSection, const FTransform& transform, FFrameTime CurrentFrameTime)
{
	auto channels = TransformSection->GetChannelProxy().GetChannels<FMovieSceneDoubleChannel>();

	for (int32 i = 0; i < channels.Num(); i++)
	{
		FMovieSceneDoubleChannel* channel = channels[i];
		FMovieSceneDoubleValue val;
		//channel[0-2] = Translation vector, channel[3-5] = Rotator, channel[6-8] = Scale Vector 
		switch (i)
		{
		case 0:
			val = FMovieSceneDoubleValue(static_cast<double>(transform.GetLocation().X));
			break;
		case 1:
			val = FMovieSceneDoubleValue(static_cast<double>(transform.GetLocation().Y));
			break;
		case 2:
			val = FMovieSceneDoubleValue(static_cast<double>(transform.GetLocation().Z));
			break;
		case 3:
			val = FMovieSceneDoubleValue(static_cast<double>(transform.Rotator().Roll));
			break;
		case 4:
			val = FMovieSceneDoubleValue(static_cast<double>(transform.Rotator().Pitch));
			break;
		case 5:
			val = FMovieSceneDoubleValue(static_cast<double>(transform.Rotator().Yaw));
			break;
		default:
			val = FMovieSceneDoubleValue(1.0);
			break;
		}
		channel->GetData().UpdateOrAddKey(CurrentFrameTime.GetFrame(), val);
	}
	TransformSection->Modify();
}


void UPeelVCamFunctionLibrary::UpdateOrAddKeysToFloatChannels(UMovieSceneFloatSection* FloatSection, bool lensKey, bool focusKey, float Lens, float FocusDistance, FFrameTime CurrentFrameTime)
{
	if (!lensKey && !focusKey) {
		return;
	}

	TArrayView<FMovieSceneFloatChannel*> Channels = FloatSection->GetChannelProxy().GetChannels<FMovieSceneFloatChannel>();

	float value = lensKey ? Lens : FocusDistance;
	Channels[0]->GetData().UpdateOrAddKey(CurrentFrameTime.GetFrame(), FMovieSceneFloatValue(value));

	FloatSection->Modify();
}

// Deletes all keys in a specified 3D Transform Section.
void UPeelVCamFunctionLibrary::DeleteAllKeysInTransformSection(UMovieScene3DTransformSection* TransformSection)
{
	if (TransformSection == nullptr)
	{
		return;
	}
	FMovieSceneChannelProxy& ChannelProxy = TransformSection->GetChannelProxy();

	// Get all float channels from the channel proxy
	TArrayView<FMovieSceneFloatChannel*> FloatChannels = ChannelProxy.GetChannels<FMovieSceneFloatChannel>();

	// Iterate over all float channels and clear them
	for (FMovieSceneFloatChannel* FloatChannel : FloatChannels)
	{
		if (FloatChannel)
		{
			FloatChannel->Reset();
		}
	}

	// Mark the section as changed to update the UI and serialization
	TransformSection->MarkAsChanged();

	// Optionally, refresh sequencer to reflect changes immediately if this is called from an editor context
	// This would require access to the Sequencer interface, which might not be available in all contexts.
}

void UPeelVCamFunctionLibrary::DeleteActorSectionsInSequence(ULevelSequence* LevelSequence, FGuid TargetActorGuid)
{
	if (!LevelSequence || !TargetActorGuid.IsValid()) return;

	UMovieScene* MovieScene = LevelSequence->GetMovieScene();
	if (!MovieScene) return;

	FGuid ActorBinding = UPeelVCamFunctionLibrary::FindActorBindingInSequence(TargetActorGuid, LevelSequence);


	if (!ActorBinding.IsValid()) return; // If no valid binding is found

	// Given the ActorBinding, we now proceed to remove the sections
	for (auto& Binding : MovieScene->GetBindings())
	{
		if (Binding.GetObjectGuid() == ActorBinding)
		{
			for (UMovieSceneTrack* Track : Binding.GetTracks())
			{
				Track->Modify(); // Mark the track as being modified before changes
				while (Track->GetAllSections().Num() > 0)
				{
					UMovieSceneSection* Section = Track->GetAllSections()[0];
					Track->RemoveSection(*Section);
				}
			}
		}
	}
	UPeelVCamFunctionLibrary::Refresh(LevelSequence);
}


FGuid UPeelVCamFunctionLibrary::FindActorBindingInSequence(FGuid TargetActorGuid, ULevelSequence* LevelSequence)
{
	if (!TargetActorGuid.IsValid() || !LevelSequence) return FGuid();

	UMovieScene* MovieScene = LevelSequence->GetMovieScene();
	if (!MovieScene) return FGuid();

	// Attempt to find a binding that matches the target actor
	for (const FMovieSceneBinding& Binding : MovieScene->GetBindings())
	{
		if (TargetActorGuid == Binding.GetObjectGuid())
		{
			return Binding.GetObjectGuid();
		}
	}

	// No matching binding found
	return FGuid();
}


void UPeelVCamFunctionLibrary::FindAdjacentTransformKeyframes(UMovieScene3DTransformSection* TransformSection, int32 CurrFrame, int32& OutPreviousKeyFrame, int32& OutNextKeyFrame)
{
	if (!TransformSection) return;
	
	// Obtain the movie scene to get its tick resolution
	UMovieScene* MovieScene = TransformSection->GetTypedOuter<UMovieScene>();
	if (!MovieScene) return;
	
	int32 PreviousClosestFrame = INT_MIN;
	int32 NextClosestFrame = INT_MAX; 
	bool bHasPreviousFrame = false, bHasNextFrame = false;

	auto channels = TransformSection->GetChannelProxy().GetChannels<FMovieSceneDoubleChannel>();

	for (FMovieSceneDoubleChannel* channel : channels)
	{
		TArrayView<const FFrameNumber> KeyTimes = channel->GetTimes();
		for (const FFrameNumber& KeyTime : KeyTimes)
		{	
			FFrameRate TickResolution = MovieScene->GetTickResolution();
			FFrameRate FrameRate = MovieScene->GetDisplayRate();

			// Calculate the frame number
			float TicksPerFrame = TickResolution.AsDecimal() / FrameRate.AsDecimal();
			int32 KeyTimeFrame = FMath::RoundToInt(KeyTime.Value / TicksPerFrame);
			if (KeyTimeFrame < CurrFrame)
			{
				if (!bHasPreviousFrame || KeyTimeFrame > PreviousClosestFrame)
				{
					PreviousClosestFrame = KeyTimeFrame;
					bHasPreviousFrame = true;
				}
			}
			else if (KeyTimeFrame > CurrFrame)
			{
				if (!bHasNextFrame || KeyTimeFrame < NextClosestFrame)
				{
					NextClosestFrame = KeyTimeFrame;
					bHasNextFrame = true;
				}
			}
		}
	}

	// Convert frame numbers back to the frame times, handling cases where no keyframe was found
	OutPreviousKeyFrame = bHasPreviousFrame ? PreviousClosestFrame : CurrFrame;
	OutNextKeyFrame = bHasNextFrame ? NextClosestFrame : CurrFrame;
}
