// Fill out your copyright notice in the Description page of Project Settings.
#include "PeelWireless.h"
#include "PeelVCam.h"
#include "Runtime/Launch/Resources/Version.h"
#include "CineCameraComponent.h"
#include "Components/SceneComponent.h"
#include <LevelSequenceActor.h>
#include <LevelSequencePlayer.h>
#include <algorithm>
#include "Kismet/GameplayStatics.h"
#include "MovieScene.h"
#include "MovieSceneSequencePlayer.h"
#include "LevelSequenceBindingReference.h"
#include "LevelSequence.h"
#include "Tracks/MovieScene3DConstraintTrack.h"
#include "Tracks/MovieScene3DTransformTrack.h"
#include "MovieSceneBinding.h"
#include "MovieSceneSection.h"
#include "Channels/MovieSceneChannelProxy.h"
#include "Sections/MovieScene3DTransformSection.h"
#include "Sections/MovieSceneFloatSection.h"
#include "IMovieScenePlayer.h"
#include "LevelSequenceEditorBlueprintLibrary.h" 
#include "MovieSceneTimeHelpers.h"
#include "Channels/MovieSceneChannelData.h"
#include "ISequencer.h"
#include "LiveLinkBlueprintLibrary.h"
#include "PeelVCamFunctionLibrary.h"

#if ENGINE_MAJOR_VERSION >= 5 && ENGINE_MINOR_VERSION >= 3
	#include "FunctionLibraries/VCamBlueprintFunctionLibrary.h"
#else 
	#include "VCamBlueprintFunctionLibrary.h"
#endif

// Sets default values for this component's properties
PEELVCAM_API UPeelWireless::UPeelWireless(FObjectInitializer const& ObjectInitializer)
	: Super(ObjectInitializer)
	, Lens(50.0)
	, FocusDistance(1000.0)
{
	PrimaryComponentTick.bCanEverTick = true;
	PrimaryComponentTick.bTickEvenWhenPaused = true;
	bWantsInitializeComponent = true;
	bTickInEditor = true;

	auto module = FModuleManager::Get().GetModulePtr<FPeelVCamModule>("PeelVCam");
	if (module)
	{
		module->Register(this);
	}

}

PEELVCAM_API UPeelWireless::~UPeelWireless()
{
	auto module = FModuleManager::Get().GetModulePtr<FPeelVCamModule>("PeelVCam");
	if (module)
	{
		module->Remove(this);
	}
}

PEELVCAM_API void UPeelWireless::InitializeComponent()
{
	Super::InitializeComponent();
	auto module = FModuleManager::Get().GetModulePtr<FPeelVCamModule>("PeelVCam");
	if (module)
	{
		module->Register(this);
	}
}
/* FIND ACTORS */
PEELVCAM_API TWeakObjectPtr<AActor> UPeelWireless::FindDollyActor()
{
	AActor* currentActor = GetOwner();
	if (currentActor == nullptr)
		return nullptr;
	FString DollyName = TEXT("Dolly");

	while (currentActor != nullptr) {
		AActor* parent = currentActor->GetAttachParentActor();
		if (parent != nullptr && parent->GetName().Contains(DollyName)) {
			return parent;
		}
		currentActor = parent;
	}
	return nullptr;
}

PEELVCAM_API TWeakObjectPtr<AActor> UPeelWireless::FindMagicCarpetActor()
{
	AActor* currentActor = GetOwner();
	if (currentActor == nullptr)
		return nullptr;
	FString MagicCarpetName = TEXT("MagicCarpet");

	while (currentActor != nullptr) {
		AActor* parent = currentActor->GetAttachParentActor();
		if (parent != nullptr && parent->GetName().Contains(MagicCarpetName)) {
			return parent;
		}
		currentActor = parent;
	}
	return nullptr;
}
PEELVCAM_API TWeakObjectPtr<ACineCameraActor> UPeelWireless::FindCineCameraActor()
{
	AActor* parent = GetOwner();
	if (parent == nullptr)
		return nullptr;

	ACineCameraActor* CineCameraActor = Cast<ACineCameraActor>(parent);

	if (CineCameraActor){
		auto CameraPtr = TWeakObjectPtr<ACineCameraActor>(CineCameraActor);
		return CameraPtr;
	}
	return nullptr;
}
/* FIND COMPONENT */
PEELVCAM_API TWeakObjectPtr<UCineCameraComponent> UPeelWireless::FindCameraComponent()
{
	AActor* parent = GetOwner();
	if (parent == nullptr)
		return nullptr;

	return parent->FindComponentByClass<UCineCameraComponent>();
}

/* LENS */
PEELVCAM_API void UPeelWireless::SetLens(float value)
{
	Lens = value;

	OnLens.Broadcast(value);

	auto Camera = FindCameraComponent();
	if (Camera != nullptr) {
		Camera->CurrentFocalLength = value;
	}
}
/* APERTURE */
PEELVCAM_API void UPeelWireless::SetAperture(float value)
{
	Aperture = value;

	OnAperture.Broadcast(value);

	auto Camera = FindCameraComponent();
	if (Camera != nullptr) {
		Camera->CurrentAperture = value;
	}
}
/* FILMBACK */
PEELVCAM_API void UPeelWireless::SetFilmback(float width, float height, float pixelAspect)
{	
	OnFilmbackChange.Broadcast();

	auto Camera = FindCameraComponent();
	if (Camera != nullptr) {
		Camera->Filmback.SensorWidth = width;
		Camera->Filmback.SensorHeight = height;
	#if ENGINE_MAJOR_VERSION >= 5 && ENGINE_MINOR_VERSION >= 2
		Camera->LensSettings.SqueezeFactor = pixelAspect;
	#endif
	}
}
/* FOCUS */
PEELVCAM_API void UPeelWireless::SetFocusDistance(float distance)
{
	FocusDistance = distance;

	OnFocus.Broadcast(distance);

	auto Camera = FindCameraComponent();
	if (Camera != nullptr) {
		Camera->FocusSettings.ManualFocusDistance = distance;
	}
}

PEELVCAM_API void UPeelWireless::SetDebugFocus(bool enable)
{
	DebugFocus = enable;

	OnDebugFocus.Broadcast(enable);

	auto Camera = FindCameraComponent();
	if (Camera != nullptr) {
		Camera->FocusSettings.bDrawDebugFocusPlane = enable;
	}
}
/* SEQUENCER */

PEELVCAM_API void UPeelWireless::DecrementFrame()
{
	OnOffsetFrame.Broadcast(-1.0);
	currentFrame += -1.0;
	UVCamBlueprintFunctionLibrary::SetCurrentLevelSequenceCurrentFrame(static_cast<int>(currentFrame));
}

PEELVCAM_API void UPeelWireless::IncrementFrame()
{
	OnOffsetFrame.Broadcast(1.0);
	currentFrame += 1.0;
	UVCamBlueprintFunctionLibrary::SetCurrentLevelSequenceCurrentFrame(static_cast<int>(currentFrame));
}

PEELVCAM_API void UPeelWireless::BeginningFrame()
{
	OnBeginningFrameEvent.Broadcast();

	ULevelSequence* Sequence = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequence();
	if (Sequence == nullptr)
		return;
	UMovieScene* MovieScene = Sequence->GetMovieScene();
	if (MovieScene == nullptr)
		return;
	FFrameRate DisplayRate = MovieScene->GetDisplayRate();
	
	int32 startFrame = ConvertFrameTime(UE::MovieScene::DiscreteInclusiveLower(MovieScene->GetPlaybackRange()), MovieScene->GetTickResolution(), DisplayRate).FloorToFrame().Value;

	UVCamBlueprintFunctionLibrary::SetCurrentLevelSequenceCurrentFrame(startFrame);

	currentFrame = static_cast<float>(startFrame);
	
}

PEELVCAM_API void UPeelWireless::SetFrame(float frameVal)
{	
	if (frameVal == 0.0) {
		OnOffsetFrame.Broadcast(frameVal);
	}
	/*OnOffsetFrame.Broadcast(frameVal);*/
	OffsetFrame = frameVal;
}

PEELVCAM_API void UPeelWireless::AddDollyTransformKey()
{
	ULevelSequence* Sequence = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequence();
	if (Sequence == nullptr)
		return;
	UMovieScene* MovieScene = Sequence->GetMovieScene();

	FFrameTime CurrentFrameTime = UPeelVCamFunctionLibrary::GetCurrentFrameTime(Sequence);

	TWeakObjectPtr<ACineCameraActor> CineCamera = FindCineCameraActor();
	if (!CineCamera.IsValid()) return;

	TWeakObjectPtr<AActor> MagicCarpet = FindMagicCarpetActor();
	if (!MagicCarpet.IsValid()) return;

	TWeakObjectPtr<AActor> Dolly = FindDollyActor();
	if (!Dolly.IsValid()) return;
	

	FGuid DollyBinding = Sequence->FindBindingFromObject(Cast<UObject>(Dolly), GetWorld());
	if (!DollyBinding.IsValid()) {
		DollyBinding = Cast<UMovieSceneSequence>(Sequence)->CreatePossessable(Cast<UObject>(Dolly));
	}
	if (!DollyBinding.IsValid()) return;
	
	UMovieScene3DTransformTrack* TransformTrack = MovieScene->FindTrack<UMovieScene3DTransformTrack>(DollyBinding);
	if (TransformTrack == nullptr) {
		TransformTrack = MovieScene->AddTrack<UMovieScene3DTransformTrack>(DollyBinding);
	}
	if (TransformTrack == nullptr) return;
	
	TArray<UMovieSceneSection*> TransformSections = TransformTrack->GetAllSections();
	UMovieScene3DTransformSection* TransformSection;
	if (TransformSections.Num() > 0) {
		TransformSection = Cast<UMovieScene3DTransformSection>(TransformSections[0]);
	}
	else {
		TransformSection = Cast<UMovieScene3DTransformSection>(TransformTrack->CreateNewSection());
		TransformTrack->AddSection(*TransformSection);
	}
	if (TransformSection == nullptr) return;
	
	if (!TransformSection->GetRange().HasLowerBound() || !TransformSection->GetRange().HasUpperBound()) {
		TransformSection->SetRange(MovieScene->GetPlaybackRange());
	}
	UPeelVCamFunctionLibrary::Refresh(Sequence);

	//Ensures any key we add will be included in the Section
	FFrameNumber sectionLowerBound = TransformSection->GetRange().GetLowerBoundValue();
	FFrameNumber sectionUpperBound = TransformSection->GetRange().GetUpperBoundValue();
	if (CurrentFrameTime.FloorToFrame().Value < sectionLowerBound.Value) {
		TransformSection->SetRange(TRange<FFrameNumber>(CurrentFrameTime.FloorToFrame(), sectionUpperBound));
	}
	if (CurrentFrameTime.FloorToFrame().Value > sectionUpperBound.Value) {
		TransformSection->SetRange(TRange<FFrameNumber>(sectionLowerBound, CurrentFrameTime.FloorToFrame()));
	}

	FTransform DollyTransform;
	FVector DollyLocation = Dolly->GetRootComponent()->GetRelativeLocation();
	FVector MagicCarpetLocation = MagicCarpet->GetRootComponent()->GetRelativeLocation();
	FVector CineCameraLocation = CineCamera->GetRootComponent()->GetRelativeLocation();
	DollyTransform.SetLocation(DollyLocation + MagicCarpetLocation + CineCameraLocation);
	
	UPeelVCamFunctionLibrary::UpdateOrAddKeysToTransformChannels(TransformSection, DollyTransform, CurrentFrameTime);

	MagicCarpet->GetRootComponent()->SetRelativeLocation(FVector::ZeroVector - CineCamera->GetRootComponent()->GetRelativeLocation());
	//CineCamera->GetRootComponent()->SetRelativeLocation(FVector::ZeroVector);
	UPeelVCamFunctionLibrary::Refresh(Sequence);
}

PEELVCAM_API void UPeelWireless::ClearDollyKeys()
{
	ULevelSequence* Sequence = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequence();
	if (Sequence == nullptr) return;

	TWeakObjectPtr<AActor> Dolly = FindDollyActor();
	if (!Dolly.IsValid()) return;

	FGuid DollyBinding = Sequence->FindBindingFromObject(Cast<UObject>(Dolly), GetWorld());
	if (!DollyBinding.IsValid()) return;

	UPeelVCamFunctionLibrary::DeleteActorSectionsInSequence(Sequence, DollyBinding);
}
PEELVCAM_API void UPeelWireless::PrevDollyKey() {

	ULevelSequence* Sequence = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequence();
	if (Sequence == nullptr)
		return;
	UMovieScene* MovieScene = Sequence->GetMovieScene();

	TWeakObjectPtr<AActor> Dolly = FindDollyActor();
	if (!Dolly.IsValid()) return;

	FGuid DollyBinding = Sequence->FindBindingFromObject(Cast<UObject>(Dolly), GetWorld());
	if (!DollyBinding.IsValid()) return;

	UMovieScene3DTransformTrack* TransformTrack = MovieScene->FindTrack<UMovieScene3DTransformTrack>(DollyBinding);
	if (TransformTrack == nullptr) return;

	TArray<UMovieSceneSection*> TransformSections = TransformTrack->GetAllSections();
	UMovieScene3DTransformSection* TransformSection;
	if (TransformSections.Num() > 0) {
		TransformSection = Cast<UMovieScene3DTransformSection>(TransformSections[0]);
	}
	else {
		return;
	}

	int32 CurrFrame = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequenceCurrentFrame();

	int32 OutPreviousKeyFrame;
	int32 OutNextKeyFrame;

	UPeelVCamFunctionLibrary::FindAdjacentTransformKeyframes(TransformSection, CurrFrame, OutPreviousKeyFrame, OutNextKeyFrame);

	UVCamBlueprintFunctionLibrary::SetCurrentLevelSequenceCurrentFrame(OutPreviousKeyFrame);

}
PEELVCAM_API void UPeelWireless::NextDollyKey() {
	ULevelSequence* Sequence = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequence();
	if (Sequence == nullptr)
		return;
	UMovieScene* MovieScene = Sequence->GetMovieScene();

	TWeakObjectPtr<AActor> Dolly = FindDollyActor();
	if (!Dolly.IsValid()) return;

	FGuid DollyBinding = Sequence->FindBindingFromObject(Cast<UObject>(Dolly), GetWorld());
	if (!DollyBinding.IsValid()) return;

	UMovieScene3DTransformTrack* TransformTrack = MovieScene->FindTrack<UMovieScene3DTransformTrack>(DollyBinding);
	if (TransformTrack == nullptr) return;

	TArray<UMovieSceneSection*> TransformSections = TransformTrack->GetAllSections();
	UMovieScene3DTransformSection* TransformSection;
	if (TransformSections.Num() > 0) {
		TransformSection = Cast<UMovieScene3DTransformSection>(TransformSections[0]);
	}
	else {
		return;
	}

	int32 CurrFrame = UVCamBlueprintFunctionLibrary::GetCurrentLevelSequenceCurrentFrame();

	int32 OutPreviousKeyFrame;
	int32 OutNextKeyFrame;

	UPeelVCamFunctionLibrary::FindAdjacentTransformKeyframes(TransformSection, CurrFrame, OutPreviousKeyFrame, OutNextKeyFrame);

	UVCamBlueprintFunctionLibrary::SetCurrentLevelSequenceCurrentFrame(OutNextKeyFrame);

}
/* LANDMARKS */
PEELVCAM_API void UPeelWireless::AddLandmark(int landmarkID) {
	TWeakObjectPtr<ACineCameraActor> CineCamera = FindCineCameraActor();
	if (!CineCamera.IsValid()) return;

	TWeakObjectPtr<AActor> MagicCarpet = FindMagicCarpetActor();
	if (!MagicCarpet.IsValid()) return;

	FLandmarkData LandmarkStruct;
	LandmarkStruct.cameraWorld = CineCamera->GetActorTransform();
	LandmarkStruct.magicCarpetWorld = MagicCarpet->GetActorTransform();
	Landmarks.Add(landmarkID, LandmarkStruct);
}

PEELVCAM_API void UPeelWireless::GotoLandmark(int landmarkID) {
	TWeakObjectPtr<ACineCameraActor> CineCamera = FindCineCameraActor();
	if (!CineCamera.IsValid()) return;

	TWeakObjectPtr<AActor> MagicCarpet = FindMagicCarpetActor();
	if (!MagicCarpet.IsValid()) return;

	FLandmarkData* LandmarkStruct = Landmarks.Find(landmarkID);

	if (LandmarkStruct == nullptr) {
		// Handle the case where the landmarkID is not found in the Landmarks map
		return;
	}
	FTransform cameraTarget = LandmarkStruct->cameraWorld;
	FTransform mcTarget = LandmarkStruct->magicCarpetWorld;

	MagicCarpet->SetActorTransform(mcTarget);

	// Calculate the difference vector from camera's current location to camera's target location
	FVector cameraDifferenceVector = cameraTarget.GetLocation() - CineCamera->GetActorTransform().GetLocation();

	// Apply the same difference to the magicCarpet
	MagicCarpet->AddActorWorldOffset(cameraDifferenceVector);
}

/* NAV */
PEELVCAM_API void UPeelWireless::TurnCameraBy(float rotation)
{
	OnTurnCameraBy.Broadcast(rotation);
	TWeakObjectPtr<ACineCameraActor> CineCamera = FindCineCameraActor();
	if (!CineCamera.IsValid()) return;

	TWeakObjectPtr<AActor> MagicCarpet = FindMagicCarpetActor();
	if (!MagicCarpet.IsValid()) return;

	FTransform CameraLocalTransform = CineCamera->GetRootComponent()->GetRelativeTransform();

	FTransform RotatorTransform = FTransform(FQuat(FRotator(0, rotation, 0)), FVector::ZeroVector, FVector::OneVector);

	FTransform MC = MagicCarpet->GetActorTransform();

	FTransform OffsetTransform = CameraLocalTransform.Inverse() * RotatorTransform * CameraLocalTransform * MC;

	MagicCarpet->SetActorTransform(OffsetTransform);
}

PEELVCAM_API void UPeelWireless::ZeroMagicCarpet()
{
	OnZeroMC.Broadcast();
	TWeakObjectPtr<ACineCameraActor> CineCamera = FindCineCameraActor();
	if (!CineCamera.IsValid()) return;

	TWeakObjectPtr<AActor> MagicCarpet = FindMagicCarpetActor();
	if (!MagicCarpet.IsValid()) return;

	MagicCarpet->GetRootComponent()->SetRelativeLocation(FVector::ZeroVector);
	MagicCarpet->GetRootComponent()->SetRelativeRotation(FRotator::ZeroRotator);

	TWeakObjectPtr<AActor> Dolly = FindDollyActor();
	if (!Dolly.IsValid()) return;

}
PEELVCAM_API void UPeelWireless::SetOffsetTranslation(FVector value)
{
	OffsetTranslation = value;
	OnOffsetTranslation.Broadcast(OffsetTranslation);

}

PEELVCAM_API void UPeelWireless::SetOffsetRotation(FVector value)
{
	OffsetRotation = value;
	OnOffsetRotation.Broadcast(OffsetRotation);
}
PEELVCAM_API void UPeelWireless::SetDampingTranslation(FVector value)
{
	DampingTranslation = value;
	OnDampingTranslation.Broadcast(DampingTranslation);
}

PEELVCAM_API void UPeelWireless::SetDampingRotation(float value)
{
	DampingRotation = value;
	OnDampingRotation.Broadcast(DampingRotation);
}
/* MISC */
PEELVCAM_API void UPeelWireless::RawButtonPressed(int buttonNum)
{
	switch (buttonNum) {
	case 1:
		OnButton1Event.Broadcast();
		break;
	case 2:
		OnButton2Event.Broadcast();
		break;
	case 3:
		OnButton3Event.Broadcast();
		break;
	case 4:
		OnButton4Event.Broadcast();
		break;
	}
}

PEELVCAM_API void UPeelWireless::RawValuePressed(int valueNum, float rawVal)
{
	switch (valueNum) {
	case 1:
		RawValue1 = rawVal;
		OnValue1Event.Broadcast(rawVal);
		break;
	case 2:
		RawValue2 = rawVal;
		OnValue2Event.Broadcast(rawVal);
		break;
	case 3:
		RawValue3 = rawVal;
		OnValue3Event.Broadcast(rawVal);
		break;
	case 4:
		RawValue4 = rawVal;
		OnValue4Event.Broadcast(rawVal);
		break;
	}
}

PEELVCAM_API void UPeelWireless::DestroyComponent(bool bPromoteChildren)
{
	Super::DestroyComponent(bPromoteChildren);
	auto module = FModuleManager::Get().GetModulePtr<FPeelVCamModule>("PeelVCam");
	if (module)
	{
		module->Register(this);
	}
}

// Called when the game starts
PEELVCAM_API void UPeelWireless::BeginPlay()
{
	Super::BeginPlay();

}

PEELVCAM_API void UPeelWireless::EndPlay(const EEndPlayReason::Type t)
{
	Super::EndPlay(t);
}

// Called every frame
PEELVCAM_API void UPeelWireless::TickComponent(float DeltaTime, ELevelTick TickType, FActorComponentTickFunction* ThisTickFunction)
{
	Super::TickComponent(DeltaTime, TickType, ThisTickFunction);
	//Translation and Rotation
	auto MagicCarpet = FindMagicCarpetActor();
	auto CineCamera = FindCineCameraActor();
	if (MagicCarpet.IsValid()){
		if (OffsetTranslation.Length() > 0.0001) {
			FTransform RelativeTransform = CineCamera->GetActorTransform();
			
			FVector RotatedOffsetTranslation = RelativeTransform.Rotator().RotateVector(OffsetTranslation);
			
			MagicCarpet->AddActorWorldOffset(RotatedOffsetTranslation);
		} 

		if (OffsetRotation.Length() > 0.0001) {

			FTransform CameraLocalTransform = CineCamera->GetRootComponent()->GetRelativeTransform();

			FTransform RotatorTransform = FTransform(FQuat(FRotator(OffsetRotation.Y, OffsetRotation.Z, OffsetRotation.X)), FVector::ZeroVector, FVector::OneVector);

			FTransform MC = MagicCarpet->GetActorTransform();

			FTransform OffsetTransform = CameraLocalTransform.Inverse() * RotatorTransform * CameraLocalTransform * MC;

			MagicCarpet->SetActorTransform(OffsetTransform);
		}
	}
	//Scrubbing
	/*if (OffsetFrame != 0.0) {
		OnOffsetFrame.Broadcast(OffsetFrame);
	}*/
	if (OffsetFrame != 0.0) {
		currentFrame += OffsetFrame;
		UVCamBlueprintFunctionLibrary::SetCurrentLevelSequenceCurrentFrame(static_cast<int>(currentFrame));
	}
	if (UVCamBlueprintFunctionLibrary::IsCurrentLevelSequencePlaying()) {
		currentFrame = static_cast<float>(UVCamBlueprintFunctionLibrary::GetCurrentLevelSequenceCurrentFrame());
	}

}

